match foo:
    case _: ...
match 1:
    case _: ...
match 1.0:
    case _: ...
match 1j:
    case _: ...
match "foo":
    case _: ...
match f"foo {x}":
    case _: ...
match {1, 2}:
    case _: ...
match ~foo:
    case _: ...
match ...:
    case _: ...
match not foo:
    case _: ...
match await foo():
    case _: ...
match lambda foo: foo:
    case _: ...
